from cython.operator cimport dereference as d
cimport ipopt as ip


cdef class IpoptApplication:
    cdef ip.SmartPtr[ip.IpoptApplication] c_app

    def __cinit__(self):
        self.c_app = ip.SmartPtr[ip.IpoptApplication](new ip.IpoptApplication())

    def initialize(self, allow_clobber=False, params_file=None):
        if params_file is not None:
            return ApplicationReturnStatus(
                d(self.c_app).Initialize(params_file, allow_clobber)
            )
        else:
            return ApplicationReturnStatus(
                d(self.c_app).Initialize(allow_clobber)
            )

    def print_copyright_message(self):
        d(self.c_app).PrintCopyrightMessage()

    def rethrow_non_ipopt_exceptions(self, dorethrow):
        return d(self.c_app).RethrowNonIpoptException(dorethrow)

    def optimize_tnlp(self, TNLP tnlp):
        cdef ip.SmartPtr[ip.TNLP] c_tnlp = new WrapperTNLP(tnlp)
        return ApplicationReturnStatus(
            d(self.c_app).OptimizeTNLP(c_tnlp)
        )

    def journalist(self):
        journalist = Journalist()
        c_journalist = d(self.c_app).Jnlst()
        if ip.IsNull(c_journalist):
            return
        journalist.c_journalist = c_journalist
        return journalist

    def options(self):
        options = OptionsList()
        c_options = d(self.c_app).Options()
        if ip.IsNull(c_options):
            return
        options.c_options = c_options
        return options

    def statistics(self):
        stats = SolveStatistics()
        c_stats = d(self.c_app).Statistics()
        if ip.IsNull(c_stats):
            return
        stats.c_statistics = c_stats
        return stats



from libcpp.string cimport string
from cython.operator cimport dereference as d
cimport pypopt.ipopt as ip
from pypopt.python_journal cimport PythonJournal as CppPythonJournal


cdef class Journal:
    cdef ip.SmartPtr[ip.Journal] c_journal

    def name(self):
        cdef string c_name = d(self.c_journal).Name()
        return c_name

    def set_print_level(self, category, level):
        cdef ip.EJournalCategory c_category = category
        cdef ip.EJournalLevel c_level = level
        d(self.c_journal).SetPrintLevel(c_category, c_level)

    def set_all_print_levels(self, level):
        cdef ip.EJournalLevel c_level = level
        d(self.c_journal).SetAllPrintLevels(c_level)

    def is_accepted(self, category, level):
        cdef ip.EJournalCategory c_category = category
        cdef ip.EJournalLevel c_level = level
        return d(self.c_journal).IsAccepted(c_category, c_level)

    def print_(self, category, level, bytes str_):
        cdef ip.EJournalCategory c_category = category
        cdef ip.EJournalLevel c_level = level
        cdef const char *c_str = str_
        d(self.c_journal).Print(c_category, c_level, c_str)

    def flush_buffer(self):
        d(self.c_journal).FlushBuffer()


cdef class PythonJournal(Journal):
    def __cinit__(self, level, out_stream):
        cdef ip.EJournalLevel c_level = level
        self.c_journal = ip.SmartPtr[ip.Journal](new CppPythonJournal(c_level, out_stream))



cdef extern from "coin/IpAlgTypes.hpp" namespace "Ipopt":
    cdef enum SolverReturn:
        SUCCESS
        MAXITER_EXCEEDED
        CPUTIME_EXCEEDED
        STOP_AT_TINY_STEP
        STOP_AT_ACCEPTABLE_POINT
        LOCAL_INFEASIBILITY
        USER_REQUESTED_STOP
        FEASIBLE_POINT_FOUND
        DIVERGING_ITERATES
        RESTORATION_FAILURE
        ERROR_IN_STEP_COMPUTATION
        INVALID_NUMBER_DETECTED
        TOO_FEW_DEGREES_OF_FREEDOM
        INVALID_OPTION
        OUT_OF_MEMORY
        INTERNAL_ERROR
        UNASSIGNED



from libcpp cimport bool
from libcpp.string cimport string
from pypopt.coin.typedef cimport Index
from pypopt.coin.smart_ptr cimport SmartPtr


cdef extern from "coin/IpJournalist.hpp" namespace "Ipopt":
    cdef cppclass Journalist:
        void Printf(EJournalLevel, EJournalCategory, const char *, ...)
        void PrintStringOverLines(EJournalLevel, EJournalCategory,
                                  Index, Index, const string&)
        void PrintfIndented(EJournalLevel, EJournalCategory, Index, const char *, ...)
        bool ProduceOutput(EJournalLevel, EJournalCategory)
        void FlushBuffer()
        bool AddJournal(const SmartPtr[Journal])
        SmartPtr[Journal] AddFileJournal(const string&, const string&, EJournalLevel)
        SmartPtr[Journal] GetJournal(const string&)
        void DeleteAllJournals()



    cdef cppclass Journal:
        string Name()
        void SetPrintLevel(EJournalCategory, EJournalLevel)
        void SetAllPrintLevels(EJournalLevel)
        bool IsAccepted(EJournalCategory, EJournalLevel)
        void Print(EJournalCategory, EJournalLevel, const char *)
        void FlushBuffer()


    cdef enum EJournalLevel:
        J_INSUPPRESSIBLE=-1,
        J_NONE=0,
        J_ERROR,
        J_STRONGWARNING,
        J_SUMMARY,
        J_WARNING,
        J_ITERSUMMARY,
        J_DETAILED,
        J_MOREDETAILED,
        J_VECTOR,
        J_MOREVECTOR,
        J_MATRIX,
        J_MOREMATRIX,
        J_ALL,
        J_LAST_LEVEL


    cdef enum EJournalCategory:
        J_DBG=0,
        J_STATISTICS,
        J_MAIN,
        J_INITIALIZATION,
        J_BARRIER_UPDATE,
        J_SOLVE_PD_SYSTEM,
        J_FRAC_TO_BOUND,
        J_LINEAR_ALGEBRA,
        J_LINE_SEARCH,
        J_HESSIAN_APPROXIMATION,
        J_SOLUTION,
        J_DOCUMENTATION,
        J_NLP,
        J_TIMING_STATISTICS,
        J_USER_APPLICATION  ,
        J_USER1  ,
        J_USER2  ,
        J_USER3  ,
        J_USER4  ,
        J_USER5  ,
        J_USER6  ,
        J_USER7  ,
        J_USER8  ,
        J_USER9  ,
        J_USER10  ,
        J_USER11  ,
        J_USER12  ,
        J_USER13  ,
        J_USER14  ,
        J_USER15  ,
        J_USER16  ,
        J_USER17  ,
        J_LAST_CATEGORY



from libcpp cimport bool
from libcpp.string cimport string
from pypopt.coin.typedef cimport Index, Number


cdef extern from "coin/IpSolveStatistics.hpp" namespace "Ipopt":
    cdef cppclass SolveStatistics:
        Index IterationCount();



import pytest
import io
from pypopt import PythonJournal, EJournalLevel, EJournalCategory
from tests.conftest import app


def test_can_add_journal(app):
    sio = io.StringIO()
    journal = PythonJournal(EJournalLevel.J_NONE, sio)
    assert app.journalist().add_journal(journal)
    assert not app.journalist().add_journal(journal)


def test_can_get_journal(app):
    sio = io.StringIO()
    journal = PythonJournal(EJournalLevel.J_NONE, sio)
    assert app.journalist().add_journal(journal)
    assert app.journalist().get_journal('PythonJournal')


def test_python_journal():
    sio = io.StringIO()
    journal = PythonJournal(EJournalLevel.J_NONE, sio)
    assert journal.name() == b'PythonJournal'
    journal.print_(EJournalCategory.J_DBG, EJournalLevel.J_NONE, b'TEST BYTES')
    assert sio.getvalue() == 'TEST BYTES'


def test_python_journal_inside_journalist(app):
    sio = io.StringIO()
    journal = PythonJournal(EJournalLevel.J_NONE, sio)
    assert app.journalist().add_journal(journal)

    app.journalist().printf(
        EJournalCategory.J_DBG, EJournalLevel.J_NONE,
        'TEST {} PLEASE IGNORE {}{}', 'POST', 0, 1
    )
    app.journalist().flush_buffer()
    assert sio.getvalue() == 'TEST POST PLEASE IGNORE 01'



from libcpp.string cimport string
from libcpp cimport bool
cimport pypopt.ipopt as ip
from pypopt.coin.journalist cimport Journal


cdef extern from "python_journal.hpp":
    cdef cppclass PythonJournal(Journal):
        PythonJournal(ip.EJournalLevel, object)



include "python_journal.pxi"
include "journalist.pxi"
include "ipopt_application.pxi"
include "options_list.pxi"
include "solve_statistics.pxi"
include "tnlp.pxi"
include "tnlp_wrapper.pxi"
include "application_return_status.pxi"



from libcpp cimport bool
from libcpp.string cimport string
from pypopt.coin.typedef cimport Index, Number


cdef extern from "coin/IpOptionsList.hpp" namespace "Ipopt":
    cdef cppclass OptionsList:
        void clear()
        void PrintList(string&)
        void PrintUserOptions(string&)

        bool SetStringValue(const string&, const string&, bool, bool)
        bool SetNumericValue(const string&, Number, bool, bool)
        bool SetIntegerValue(const string&, Index, bool, bool)

        bool SetStringValueIfUnset(const string&, const string&, bool, bool)
        bool SetNumericValueIfUnset(const string&, Number, bool, bool)
        bool SetIntegerValueIfUnset(const string&, Index, bool, bool)

        bool GetStringValue(const string&, string&, const string&)
        bool GetEnumValue(const string&, Index&, const string&)
        bool GetBoolValue(const string&, bool&, const string&)
        bool GetNumericValue(const string&, Number&, const string&)
        bool GetIntegerValue(const string&, Index&, const string&)



# Copyright 2018 Francesco Ceccon
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

cimport pypopt.ipopt as ip


cdef cppclass WrapperTNLP(ip.TNLP):
    TNLP owner

    WrapperTNLP(TNLP owner_):
        this.owner = owner_

    bool get_nlp_info(ip.Index &n, ip.Index &m, ip.Index &nnz_jac_g,
                      ip.Index &nnz_h_lag, ip.IndexStyleEnum &index_style):
        info = owner.get_nlp_info()
        if info is None:
            return False
        (&n)[0] = info.n
        (&m)[0] = info.m
        (&nnz_jac_g)[0] = info.nnz_jac
        (&nnz_h_lag)[0] = info.nnz_hess
        (&index_style)[0] = info.index_style
        return True

    bool get_bounds_info(ip.Index n, ip.Number *x_l, ip.Number *x_u,
                         ip.Index m, ip.Number *g_l, ip.Number *g_u):
        cdef ip.Number[:] py_x_l
        cdef ip.Number[:] py_x_u
        cdef ip.Number[:] py_g_l
        cdef ip.Number[:] py_g_u

        if n:
            py_x_l = <ip.Number[:n]>x_l
            py_x_u = <ip.Number[:n]>x_u
        else:
            py_x_l = None
            py_x_u = None

        if m:
            py_g_l = <ip.Number[:m]>g_l
            py_g_u = <ip.Number[:m]>g_u
        else:
            py_g_l = None
            py_g_u = None

        return owner.get_bounds_info(py_x_l, py_x_u, py_g_l, py_g_u)

    bool get_starting_point(ip.Index n, bool init_x, ip.Number *x,
                            bool init_z, ip.Number *z_l, ip.Number *z_u,
                            ip.Index m, bool init_lambda, ip.Number *lambda_):
        cdef ip.Number[:] py_x
        cdef ip.Number[:] py_z_l
        cdef ip.Number[:] py_z_u
        cdef ip.Number[:] py_lambda

        if init_x:
            py_x = <ip.Number[:n]>x
        else:
            py_x = None

        if init_z:
            py_z_l = <ip.Number[:n]>z_l
            py_z_u = <ip.Number[:n]>z_u
        else:
            py_z_l = py_z_u = None

        if init_lambda:
            py_lambda = <ip.Number[:m]>lambda_
        else:
            py_lambda = None

        return owner.get_starting_point(
            init_x, py_x, init_z, py_z_l, py_z_u, init_lambda, py_lambda
        )

    bool eval_f(ip.Index n, const ip.Number *x, bool new_x, ip.Number &obj_val):
        cdef ip.Number[:] py_x = <ip.Number[:n]>x
        py_obj_val = owner.eval_f(py_x, new_x)
        (&obj_val)[0] = py_obj_val
        return True

    bool eval_grad_f(ip.Index n, const ip.Number *x, bool new_x,
                     ip.Number *grad_f):
        cdef ip.Number[:] py_x = <ip.Number[:n]>x
        cdef ip.Number[:] py_grad_f = <ip.Number[:n]>grad_f
        return owner.eval_grad_f(py_x, new_x, py_grad_f)

    bool eval_g(ip.Index n, const ip.Number *x, bool new_x, ip.Index m,
                ip.Number *g):
        cdef ip.Number[:] py_x = <ip.Number[:n]>x
        cdef ip.Number[:] py_g
        if m:
            py_g = <ip.Number[:m]>g
        else:
            py_g = None
        return owner.eval_g(py_x, new_x, py_g)

    bool eval_jac_g(ip.Index n, const ip.Number *x, bool new_x, ip.Index m,
                    ip.Index nele_jac, ip.Index *row, ip.Index *col,
                    ip.Number *values):
        cdef ip.Index[:] py_row
        cdef ip.Index[:] py_col
        cdef ip.Number[:] py_x
        cdef ip.Number[:] py_values

        if values == NULL:
            if nele_jac:
                py_row = <ip.Index[:nele_jac]>row
                py_col = <ip.Index[:nele_jac]>col
            else:
                py_row = py_col = None
            return owner.get_jac_g_structure(py_row, py_col)
        else:
            py_x = <ip.Number[:n]>x
            if nele_jac:
                py_values = <ip.Number[:nele_jac]>values
            else:
                py_values = None
            return owner.eval_jac_g(py_x, new_x, py_values)

    bool eval_h(ip.Index n, const ip.Number *x, bool new_x, ip.Number
                obj_factor, ip.Index m, const ip.Number *lambda_,
                bool new_lambda, ip.Index nele_hess, ip.Index *row,
                ip.Index *col, ip.Number *values):
        cdef ip.Index[:] py_row
        cdef ip.Index[:] py_col
        cdef ip.Number[:] py_x
        cdef ip.Number[:] py_values
        cdef ip.Number[:] py_lambda

        if values == NULL:
            py_row = <ip.Index[:nele_hess]>row
            py_col = <ip.Index[:nele_hess]>col
            return owner.get_h_structure(py_row, py_col)
        else:
            py_x = <ip.Number[:n]>x
            if m:
                py_lambda = <ip.Number[:m]>lambda_
            else:
                py_lambda = None
            py_values = <ip.Number[:nele_hess]>values
            return owner.eval_h(
                py_x, new_x, obj_factor, py_lambda, new_lambda, py_values
            )

    void finalize_solution(ip.SolverReturn status, ip.Index n,
                           const ip.Number *x, const ip.Number *z_l,
                           const ip.Number *z_u, ip.Index m, const ip.Number *g,
                           const ip.Number *lambda_, ip.Number obj_value,
                           const ip.IpoptData *ip_data,
                           ip.IpoptCalculatedQuantities *ip_cq):
        cdef ip.Number[:] py_x = <ip.Number[:n]>x
        cdef ip.Number[:] py_z_l = <ip.Number[:n]>z_l
        cdef ip.Number[:] py_z_u = <ip.Number[:n]>z_u
        cdef ip.Number[:] py_g
        cdef ip.Number[:] py_lambda

        if m:
            py_g = <ip.Number[:m]>g
            py_lambda = <ip.Number[:m]>lambda_
        else:
            py_g = None
            py_lambda = None

        owner.finalize_solution(
            py_x, py_z_l, py_z_u, py_g, py_lambda, obj_value
        )
